/* Tower Toppler - Nebulus
 * Copyright (C) 2000-2006  Andreas Rver
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.

 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.

 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

#include "highscore.h"
#include "decl.h"
#include "screen.h"

#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include <fcntl.h>
#include <unistd.h>

#ifndef WIN32
   #include <sys/stat.h>
   #include <sys/file.h>
   #include <errno.h>
#endif

#ifdef __QNXNTO__
   #include <strings.h>
#endif // __QNXNTO__

#ifdef XCODE
   #include "osx.h"
#endif

#define NUMHISCORES 10

#define SCOREFNAME "toppler.hsc"

typedef struct {
  Uint32 points;
  char name[SCORENAMELEN+1];
  Sint16 tower; /* tower reached, -1 = mission finished */
} _scores;

static _scores scores[NUMHISCORES];

/* this is the name of the surrenlty selected mission */
static char missionname[100];

/* File pointer to the highscore file, we keep this open till the game exits,
   as we need special rights to open it, and drop those immediately after
   opening it. */
static FILE *scores_f = NULL;

static void savescores(void) {

  unsigned char len;
  char mname[256];
  
  if (scores_f == NULL)
    return;

  fseek(scores_f, 0, SEEK_SET);
  while (!feof(scores_f)) {

    if ((fread(&len, 1, 1, scores_f) == 1) &&
        (fread(mname, 1, len, scores_f) == len)) {
      mname[len] = 0;
      if (strcasecmp(mname, missionname) == 0) {

        // this is necessary because some system can not switch
        // on the fly from reading to writing
        fseek(scores_f, ftell(scores_f), SEEK_SET);

        fwrite(scores, sizeof(_scores)*NUMHISCORES, 1, scores_f);
        return;
      }
    } else
      break;

    fseek(scores_f, ftell(scores_f) + sizeof(_scores)*NUMHISCORES, SEEK_SET);
  }

  unsigned char tmp = strlen(missionname);

  fwrite(&tmp, 1, 1, scores_f);
  fwrite(missionname, 1, tmp, scores_f);
  fwrite(scores, sizeof(_scores)*NUMHISCORES, 1, scores_f);
}

static void loadscores(void) {

  unsigned char len;
  char mname[256];
  
  if (scores_f == NULL)
    return;

  fseek(scores_f, 0, SEEK_SET);
  while (!feof(scores_f)) {

    if ((fread(&len, 1, 1, scores_f) == 1) &&
        (fread(mname, 1, len, scores_f) == len) &&
        (fread(scores, 1, sizeof(_scores) * NUMHISCORES, scores_f) == sizeof(_scores) * NUMHISCORES)) {
      mname[len] = 0;
      if (strcasecmp(mname, missionname) == 0)
        return;
    }
  }

  for (int t = 0; t < NUMHISCORES; t++) {
    scores[t].points = 0;
    scores[t].name[0] = 0;
  }
}

static bool hsc_lock(void) {

#ifndef WIN32

  if (scores_f == NULL)
    return false;

  while (1) {  
    int res = flock(fileno(scores_f), LOCK_EX);
    if (res) {
      if (errno == EINTR)
        continue;
      perror("Error locking highscore file");
      return false;
    }
    return true;
  }

#endif

  return true;
}

static void hsc_unlock(void) {

#ifndef WIN32

  flock(fileno(scores_f), LOCK_UN);

#endif
}

void hsc_init(void) {

  for (int t = 0; t < NUMHISCORES; t++) {
    scores[t].points = 0;
    scores[t].name[0] = 0;
    scores[t].tower = 0;
  }

#ifndef WIN32

#ifdef HISCOREDIR

  /* Try to open the global scoreboard file, while we still have the rights */
  scores_f = fopen(HISCOREDIR "/" SCOREFNAME, "r+");
  /* If we cannot open it, try creating it */
  if (!scores_f)
    scores_f = fopen(HISCOREDIR "/" SCOREFNAME, "w+");

  /* Figure out who we really are. */
  gid_t realgid = getgid();
  uid_t realuid = getuid();

  /* This is where we drop our setuid/setgid privileges. */
  if (setresgid(-1, realgid, realgid) != 0) {
    perror("Could not drop setgid privileges.  Aborting.");
    exit(1);
  }
  if (setresuid(-1, realuid, realuid) != 0) {
    perror("Could not drop setuid privileges.  Aborting.");
    exit(1);
  }

#endif

  /* Ensure $HOME/.toppler exists */
  char name[PATH_MAX];

  snprintf(name, sizeof(name), "%s/.toppler", homedir());
  if (access(name, F_OK) != 0)
    mkdir(name, 0775);

  /* Did opening the global highscore file succeed? */
  if (scores_f)
    return; /* All done */

  /* Fall back to using a highscore file in $HOME/.toppler */
  snprintf(name, sizeof(name), "%s/.toppler/%s", homedir(), SCOREFNAME);
  scores_f = fopen(name, "r+");
  if (!scores_f)
    scores_f = fopen(name, "w+");
  if (scores_f)
    return; /* All done */

#endif

  /* Fall back to using a highscore file in the cwd */
  scores_f = fopen(SCOREFNAME, "r+");
  if (!scores_f)
    scores_f = fopen(SCOREFNAME, "w+");
}

void hsc_cleanup(void)
{
  if (scores_f) {
    fclose(scores_f);
    scores_f = NULL;
  }
}

void hsc_select(const char * mission) {
  snprintf(missionname, sizeof(missionname), "%s", mission);
  bool locked = hsc_lock();
  loadscores();
  if (locked)
    hsc_unlock();
}

Uint8 hsc_entries(void) { return NUMHISCORES; }

void hsc_entry(Uint8 nr, char *name, Uint32 *points, Uint8 *tower) {

  if (nr < NUMHISCORES) {
    if (name) strcpy(name, scores[nr].name);
    if (points) *points = scores[nr].points;
    if (tower) *tower = scores[nr].tower;
  } else {
    if (name) name[0] = 0;
    if (points) *points = 0;
    if (tower) *tower = 0;
  }
}

bool hsc_canEnter(Uint32 points) {
  return points > scores[NUMHISCORES-1].points;
}

Uint8 hsc_enter(Uint32 points, Uint8 tower, char *name) {

    int t = NUMHISCORES;
    
    bool locked = hsc_lock();

    loadscores();
      
    while ((t > 0) && (points > scores[t-1].points)) {
      if (t < NUMHISCORES)
        scores[t] = scores[t-1];
      t--;
    }
  
    if (t < NUMHISCORES) {
  
      snprintf(scores[t].name, sizeof(scores[t].name), "%s", name);
      scores[t].points = points;
      scores[t].tower = tower;

      savescores();
    } else {
      t = 0xff;
    }

    if (locked)
      hsc_unlock();

    return t;
}

